-- 使用 instance entity mesh，模拟 GPU compute shadder 的工作组如何分配工作项
-- 彩色版本：用不同颜色区分各个本地工作组
---- Codea 中的 buffer 名字为 instPosBuffer，对应的 compute shader  中也必须使用 instPos 这个名字
---- 使用 append() 方法为 buffer 追加元素很容易越界
---- 分别阐述 1，2，3 维的本地工作组，以及 1，2，3 维的全局工作组组合起来的多种布局场景
---- 分别阐述计算总任务数(buffer总数)和全局工作组尺寸不同关系时的任务分配情况： 小于，等于，大于(此时会使用多个全局工作组)
---- 可以用6个 mesh 和 6个 compute shader 来动态演示不同的数组寻址效果(1维，2维，3维)
---- 高维数组也可以视觉化演示，不过同时最多只能展示3个维度


---[=[

instancing = shader {
    name = "instancing",
    
    properties =
    {
        {"instPosBuffer", "buffer"},
        {"instColBuffer", "buffer"},
        {"utexCoorOffset", "float", 0.01},
        {"utintColor", "vec4"},
        {"sTexture", "texture"},       
    },
    
    pass =
    {
        --[[
        cullFace = "none", -- backgrounds don't need culling
        depthWrite = false, -- no need to write to depth buffer
        depthFunc = "always", -- no need for depth testing
        blendMode = "disabled", -- no need for blending
        --renderQueue = "background", -- render behind everything else
        --]]
        
        
        renderQueue = OPAQUE,
        blendMode = DISABLED,
        depthWrite = DEPTH_WRITE_ENABLED,
        depthFunc = DEPTH_FUNC_LESS_EQUAL,
        
        vertex =
        [[
        #version 430
        #include <codea/common.glsl>
        
        layout (location = POSITION) in vec3 a_position;
        layout (location = TEXCOORD0) in vec2 a_texcoord0;
        
        out vec2 uv;
        out vec4 col;
        
        // 利用 instances buffer 接收 compute shader 的计算结果
        struct Instance
        {
        vec4 position;
        };
        BUFFER_RW(instPos, Instance, BUFFER0);
        
        struct Color
        {
        vec4 color0;
        };
        BUFFER_RW(instCol, Color, BUFFER1);
        
        void main()
        {
        vec3 posOffset = instPos[gl_InstanceID].position.xyz;
        vec3 pos = a_position + posOffset;
        //pos = pos + posOffset;
        
        vec4 lwCol = instCol[gl_InstanceID].color0;
        //col = normalize(pos);
        col = lwCol;
        
        gl_Position = u_modelViewProj * vec4(pos, 1.0);
        //gl_Position = u_modelViewProj * vec4(posOffset, 1.0);
        //gl_Position = u_modelViewProj * vec4(a_position + vec3(instPos[gl_InstanceID].position.xyz), 1.0);
        //gl_Position = u_modelViewProj * vec4(a_position, 1.0);
        uv = a_texcoord0;
        }
        ]],
        
        fragment = [[
        #version 430
        #include <codea/common.glsl>
        
        in vec2 uv;
        in vec4 col;
        out vec4 FragColor;
        
        uniform sampler2D sTexture;//纹理内容数据
        uniform float utexCoorOffset;//纹理坐标偏移
        uniform vec4 utintColor;
        
        void main()
        {
        //计算纹理坐标的偏移      
        float oy = sin(utexCoorOffset)*1;
        vec2 texCoorTemp=fract(uv + vec2(utexCoorOffset, oy));
        //给此片元从纹理中采样出颜色值      
        vec4 finalColor=texture(sTexture, texCoorTemp); 
        //vec4 finalColor = texture(sTexture, uv); 
        //FragColor = vec4(col.rgb, 1.0);
        FragColor = finalColor * vec4(col.rgb, 1.0);
        //FragColor = finalColor * col;
        //FragColor = finalColor*vec4(col.rgb, 1.0)*utintColor;
        }
        ]],
    }
}

function setup()
    -- 使用基本场景的默认镜头
    scn = scene.default3d()
    scn.sky = nil
    -- scn.camera.farPlane =5000
    
    cam = scn.camera:get(camera)
    cam.farPlane = 50000
    
    ---[[ create a sun entity and add a directional light to it
    sunEntity = scn:entity("sun")
    sun = sunEntity:add(light.directional(vec3(-1,-1,-1)))
    sun.color = color(233, 150, 80)
    -- sun.clearColor = color(233, 150, 80)
    sun.intensity = 0.05   
    -- sun.direction =vec3(-100,-100,-100)
    sun.castShadows = true
    --]]
    
    
    local rig = scn.camera:add(camera.rigs.orbit)
    rig.distance = 100
    rig.angles.x = 45    
    rig.angles.y = -75
    
    -- 地面
    ground = scn:entity("ground")
    ground:add(mesh.box(50, 1, 25))
    -- ground:add(physics3d.body, STATIC):box(50,1,50)
    -- (this will work in the next build)
    ground.material.map = asset.documents.Dropbox.n0
    ground.material.normalMap = asset.documents.Dropbox.n2
    -- ground.material.occlusionMap= asset.documents.Dropbox.dd
    ground.y = -1.5
    
    -- 新建 material
    checkers = material.unlit()
    checkers.map = image.read(asset.documents.checker_png)
    checkers.color = color(79, 226, 18)
    checkers.scaleOffset = vec4(0.125, 0.125, 0.0, 0.0)
    checkers.roughness = 0.5
    -- checkers.shadowCascade.size = 1
    
    
    -- Create a buffer to match the shader buffer (syntax is a bit ugly right now)
    local layout = shader.bufferLayout { {shader.position, 4, shader.float} }
    -- Create buffer to hold instance data
    instPosBuffer = shader.buffer(layout)
    
    -- Create a buffer to match the shader buffer (syntax is a bit ugly right now)
    local layout = shader.bufferLayout { {shader.color0, 4, shader.float} }
    -- Create buffer to hold instance data
    instColBuffer = shader.buffer(layout)
    
    initBuffer()
    
    VisualColorWorkGroup.ulx =1
    VisualColorWorkGroup:setBuffer("instPosBuffer", instPosBuffer, shader.readwrite)
    VisualColorWorkGroup:setBuffer("instColBuffer", instColBuffer, shader.readwrite)
    VisualColorWorkGroup.instanceCount = instPosBuffer.size
    
    instancing:setBuffer("instPosBuffer", instPosBuffer, shader.readwrite)
    instancing:setBuffer("instColBuffer", instColBuffer, shader.readwrite)
    
    print("buffer size: ", instPosBuffer.size)
    print("compute group size: ", VisualColorWorkGroup:workgroupSize())
    
    
    texCoorOffset=0.0
    instancing.sTexture = image.read(asset.builtin.Blocks.Brick_Red)
    instancing.utexCoorOffset = 1.5
    instancing.utintColor = color(243, 244, 243)
    
    --[[ create a sphere mesh
    msh = mesh.sphere(1.3,32,16,0,360,0,180)
    -- msh.material = checkers
    msh.shader = instancing    
    
    -- 新建entity mesh
    eMesh = scn:entity("eMesh")
    eMesh.mesh = mesh.sphere(1.3,32,16,0,360,0,180)
    eMesh.postition = vec3(-10,20,10)
    eMesh.mesh = mesh.read(asset.builtin.Primitives.Monkey_obj)
    eMesh.mesh.shader = instancing
    --]]
    
    -- 新建一个使用 updatePositions 作为 compute shader 的 mesh
    eMesh1 = scn:entity("eMesh1")
    -- eMesh1.mesh = mesh.sphere(6.3,32,16,0,360,0,180)
    eMesh1.mesh = mesh.box(vec3(1,1,1), vec3(14,4,4))
    eMesh1.postition = vec3(0,0,0)
    -- eMesh1.mesh = mesh.read(asset.builtin.Blocky_Characters.Adventurer)
    eMesh1.mesh.shader = instancing
    eMesh1.active = false
    
    parameter.action("clear buffer", function() instPosBuffer:clear() end)
    parameter.action("init buffer", function() initBuffer() end)
    parameter.vec3("colR",vec3(0.9,0.9,0.9))
    parameter.integer("factor",2,4,2)
    VisualColorWorkGroup.ulx =1
    VisualColorWorkGroup.factor =2
    VisualColorWorkGroup.colR =vec3(0.9,0.9,0.9)

    -- 延时动画演示实例绘制顺序：分别在 xyz 三个维度方向上展开
    -- scn:tween(eMesh1):to{worldRotation = eMesh1.worldRotation * quat.angleAxis(145, vec3(0,1,0))}:time(2.25):ease(tween.backOut)            
    
    scene.main = scn
end

function initBuffer()
    -- Create initial instances by appending data to buffer
    for k = 1, 6400 do
        -- instancesBuffer:append(shader.position, vec4(math.random(0, WIDTH/2),math.random(0, HEIGHT/2),math.random(0, WIDTH/2),1))
        instPosBuffer:append(shader.position, vec4(0.0, 0.0, 0.0, 1.0)) 
        instColBuffer:append(shader.color0, vec4(0.0, 0.0, 0.0, 1.0)) 
    end
end


function draw()
    background(0)
    VisualColorWorkGroup.ulx =1
    VisualColorWorkGroup.factor = factor or 2
    VisualColorWorkGroup.colR = colR or vec3(0.5,0.5,0.5)
    ---[[ 根据工作组总数(所有工作项，也就是线程的数目) workgroupSize，来分配任务到 GPU 上
    local wx,wy,wz = VisualColorWorkGroup:workgroupSize()
    -- VisualWorkGroup:dispatch(instancesBuffer.size/(wx*wy*wz), 1, 1)
    VisualColorWorkGroup:dispatch(5,6,7)
    --]]    
    
    texCoorOffset = (texCoorOffset+0.01)%10.0;
    instancing.utexCoorOffset = texCoorOffset
    
    scn:draw()
    matrix.push()
    -- camera.perspective(45, 0.1, 1000)
    -- matrix.translate(WIDTH/2, HEIGHT/2, 600).scale(10,10,10)
    -- msh:draw(instPosBuffer.size)
    -- eMesh.mesh:draw(instPosBuffer.size/100)
    eMesh1.mesh:draw(instPosBuffer.size)
    -- 动态演示实例绘制顺序
    -- eMesh1.mesh:draw(math.floor(time.elapsed*10))

    matrix.pop()
end

--[[
function scn:entity("customDrawing"):draw()
    myMesh:draw(instanceCount)
end
--]]

function touched(touch)
    if touch.state == BEGAN then
        local wx,wy,wz = VisualColorWorkGroup:workgroupSize()
        print("workgroupSize: ", wx,wy,wz)
        -- instancesBuffer:clear()
    end
end

--]=]
